@@ -101,6 +101,7 @@ normal()
 	int				ctrl_w = FALSE;			/* got CTRL-W command */
 	int				old_col = 0;
 	int				dont_adjust_op_end = FALSE;
+	static int		search_dont_set_mark = FALSE;	/* for "*" and "#" */
 
 	Prenum = 0;
 	/*
@@ -135,10 +136,20 @@ getcount:
 			Prenum = Prenum * 10 + (c - '0');
 		if (Prenum < 0)			/* got too large! */
 			Prenum = 999999999;
+		if (ctrl_w)
+		{
+			++no_mapping;
+			++allow_keys;				/* no mapping for nchar, but keys */
+		}
 		c = vgetc();
 #ifdef HAVE_LANGMAP
 		LANGMAP_ADJUST(c, TRUE);
 #endif
+		if (ctrl_w)
+		{
+			--no_mapping;
+			--allow_keys;
+		}
 		(void)add_to_showcmd(c, FALSE);
 	}
 
@@ -287,11 +298,13 @@ getcount:
 	  case Ctrl('B'):
 	  case K_S_UP:
 	  case K_PAGEUP:
+	  case K_KPAGEUP:
 		dir = BACKWARD;
 
 	  case Ctrl('F'):
 	  case K_S_DOWN:
 	  case K_PAGEDOWN:
+	  case K_KPAGEDOWN:
 		if (checkclearop())
 			break;
 		(void)onepage(dir, Prenum1);
@@ -529,17 +542,6 @@ dozet:
 			break;
 			/* print full name if count given or :cd used */
 		fileinfo(did_cd | (int)Prenum, FALSE, FALSE);
-
-		/*
-		 * In Visual mode and "^O^G" in Insert mode, the message will be
-		 * overwritten by the mode message.  Wait a bit, until a key is hit.
-		 */
-		if ((VIsual_active || (restart_edit && p_smd)) && KeyTyped)
-		{
-			setcursor();
-			flushbuf();
-			mch_delay(10000L, FALSE);
-		}
 		break;
 
 	  case K_CCIRCM:			/* CTRL-^, short for ":e #" */
@@ -548,15 +550,19 @@ dozet:
 		(void)buflist_getfile((int)Prenum, (linenr_t)0, GETF_SETMARK|GETF_ALT);
 		break;
 
-	  case 'Z': 		/* write, if changed, and exit */
+		/*
+		 * "ZZ": write if changed, and exit window
+		 * "ZQ": quit window (Elvis compatible)
+		 */
+	  case 'Z':
 		if (checkclearopq())
 			break;
-		if (nchar != 'Z')
-		{
+		if (nchar == 'Z')
+			stuffReadbuff((char_u *)":x\n");
+		else if (nchar == 'Q')
+			stuffReadbuff((char_u *)":q!\n");
+		else
 			clearopbeep();
-			break;
-		}
-		stuffReadbuff((char_u *)":x\n");
 		break;
 
 	  case Ctrl(']'):			/* :ta to current identifier */
@@ -619,9 +625,14 @@ search_word:
 					stuffReadbuff((char_u *)"?");
 
 				/*
-				 * put cursor at start of word, makes search skip the word
-				 * under the cursor
+				 * Put cursor at start of word, makes search skip the word
+				 * under the cursor.
+				 * Call setpcmark() first, so "*``" puts the cursor back where
+				 * it was, and set search_dont_set_mark to avoid doing it
+				 * again when searching.
 				 */
+				setpcmark();
+				search_dont_set_mark = TRUE;
 				curwin->w_cursor.col = ptr - ml_get_curline();
 
 				if (c != 'g' && iswordchar(*ptr))
@@ -873,6 +884,7 @@ lineop:
 		break;
 
 	  case K_HOME:
+	  case K_KHOME:
 		if ((mod_mask & MOD_MASK_CTRL))
 			goto goto_line_one;
 		Prenum = 1;
@@ -956,6 +968,7 @@ dowrdcmd:
 		break;
 
 	  case K_END:
+	  case K_KEND:
 		if ((mod_mask & MOD_MASK_CTRL))
 			goto goto_line;
 		/* FALLTHROUGH */
@@ -996,11 +1009,13 @@ dowrdcmd:
 		curwin->w_set_curswant = TRUE;
 
 		n = do_search(c, searchbuff, Prenum1,
-						 SEARCH_MARK | SEARCH_OPT | SEARCH_ECHO | SEARCH_MSG);
+				(search_dont_set_mark ? 0 : SEARCH_MARK) |
+				SEARCH_OPT | SEARCH_ECHO | SEARCH_MSG);
 		if (n == 0)
 			clearop();
 		else if (n == 2)
 			op_motion_type = MLINE;
+		search_dont_set_mark = FALSE;
 		break;
 
 	  case 'N':
@@ -1911,6 +1926,7 @@ cursormark:
 
 			case '0':
 			case K_HOME:
+			case K_KHOME:
 				op_motion_type = MCHAR;
 				op_inclusive = FALSE;
 				if (curwin->w_p_wrap)
@@ -1931,6 +1947,7 @@ cursormark:
 
 			case '$':
 			case K_END:
+			case K_KEND:
 				op_motion_type = MCHAR;
 				op_inclusive = TRUE;
 				if (curwin->w_p_wrap)
@@ -2003,8 +2020,8 @@ gotofile:
 					if (curbuf->b_changed && curbuf->b_nwindows <= 1 && !p_hid)
 						autowrite(curbuf);
 					setpcmark();
-					(void)do_ecmd(0, ptr, NULL, NULL, p_hid, (linenr_t)0,
-																	   FALSE);
+					(void)do_ecmd(0, ptr, NULL, NULL, (linenr_t)0,
+													   p_hid ? ECMD_HIDE : 0);
 					vim_free(ptr);
 				}
 				else
@@ -2177,6 +2194,44 @@ goto_line_one:
 	do_pending_operator(c, nchar, finish_op, searchbuff, 
 					  &command_busy, old_col, FALSE, dont_adjust_op_end);
 
+	/*
+	 * Wait when a message is displayed that will be overwritten by the mode
+	 * message.
+	 * In Visual mode and with "^O" in Insert mode, a short message will be
+	 * overwritten by the mode message.  Wait a bit, until a key is hit.
+	 * Only do this if the command was typed, not from a mapping.
+	 * Also wait a bit after an error message, e.g. for "^O:".
+	 * Don't redraw the screen, it would remove the message.
+	 */
+	if (((p_smd && (VIsual_active || restart_edit) &&
+			(clear_cmdline || redraw_cmdline) && msg_didany && KeyTyped) ||
+			(restart_edit && !VIsual_active && (msg_scroll || emsg_on_display
+#ifdef SLEEP_IN_EMSG
+										|| need_sleep
+#endif
+														))) &&
+			yankbuffer == 0 && !command_busy && stuff_empty() && op_type == NOP)
+	{
+		++RedrawingDisabled;
+		cursupdate();
+		--RedrawingDisabled;
+		setcursor();
+		flushbuf();
+		if (msg_scroll || emsg_on_display
+#ifdef SLEEP_IN_EMSG
+											|| need_sleep
+#endif
+															)
+			mch_delay(1000L, TRUE);		/* wait at least one second */
+		mch_delay(10000L, FALSE);		/* wait up to ten seconds */
+
+		msg_scroll = FALSE;
+		emsg_on_display = FALSE;
+#ifdef SLEEP_IN_EMSG
+		need_sleep = FALSE;
+#endif
+	}
+
 normal_end:
 	if (op_type == NOP && yankbuffer == 0)
 		clear_showcmd();
@@ -2185,7 +2240,8 @@ normal_end:
 						 && !command_busy && stuff_empty() && yankbuffer == 0)
 		(void)edit(restart_edit, FALSE, 1L);
 
-	checkpcmark();			/* check if we moved since setting pcmark */
+	if (!search_dont_set_mark)
+		checkpcmark();			/* check if we moved since setting pcmark */
 	vim_free(searchbuff);
 
 /*
