@@ -874,8 +874,6 @@ mch_settitle(title, icon)
 	/*
 	 * Note: if terminal is xterm, title is set with escape sequence rather
 	 * 		 than x11 calls, because the x11 calls don't always work
-	 * Check only if the start of the terminal name is "xterm", also catch
-	 * "xterms".
 	 */
 	if (is_xterm(term_strings[KS_NAME]))
 		type = 2;
@@ -1373,6 +1371,9 @@ mch_settmode(raw)
 									/* but it breaks function keys on MINT */
 # endif
 								);
+# ifdef ONLCR		/* don't map NL -> CR NL, we do it ourselves */
+		tnew.c_oflag &= ~ONLCR;
+# endif
 		tnew.c_cc[VMIN] = 1;			/* return after 1 char */
 		tnew.c_cc[VTIME] = 0;			/* don't wait */
 # if defined(HAVE_TERMIOS_H)
@@ -1685,8 +1686,9 @@ call_shell(cmd, options)
 		x = system(p_sh);
 	else
 	{
-		sprintf(newcmd, "%s %s -c \"%s\"", p_sh,
+		sprintf(newcmd, "%s %s %s \"%s\"", p_sh,
 					extra_shell_arg == NULL ? "" : (char *)extra_shell_arg,
+					(char *)p_shcf,
 					(char *)cmd);
 		x = system(newcmd);
 	}
@@ -1786,7 +1788,7 @@ call_shell(cmd, options)
 	{
 		if (extra_shell_arg != NULL)
 			argv[argc++] = (char *)extra_shell_arg;
-		argv[argc++] = "-c";
+		argv[argc++] = (char *)p_shcf;
 		argv[argc++] = (char *)cmd;
 	}
 	argv[argc] = NULL;
@@ -2060,14 +2062,8 @@ call_shell(cmd, options)
 						{
 #ifdef SIGINT
 							if (buffer[0] == Ctrl('C'))
-							{
-								/* Use both kill() and killpg(), in case one
-								 * of the two fails */
-								kill(pid, SIGINT);
-# ifdef HAVE_KILLPG
-								killpg(0, SIGINT);
-# endif
-							}
+								/* send SIGINT to all processes in our group */
+								kill(0, SIGINT);
 #endif
 							if (pty_master_fd < 0 && toshell_fd >= 0 &&
 													   buffer[0] == Ctrl('D'))
@@ -2444,7 +2440,6 @@ RealWaitForChar(fd, msec)
  * list_notfound is ignored
  */
 
-extern char *mktemp __ARGS((char *));
 #ifndef SEEK_SET
 # define SEEK_SET 0
 #endif
@@ -2473,11 +2468,8 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 	files_alloced = EXPL_ALLOC_INC;	/* how much space is allocated */
 	files_free = EXPL_ALLOC_INC;	/* how much space is not used  */
 	*file = (char_u **) alloc(sizeof(char_u **) * files_alloced);
-	if (!*file)
-	{
-		emsg(e_outofmem);
+	if (*file == NULL)
 		return FAIL;
-	}
 
 	for (; num_pat > 0; num_pat--, pat++)
 	{
@@ -2485,17 +2477,12 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 		if (vim_strchr(*pat, '$') || vim_strchr(*pat, '~'))
 		{
 			/* expand environment var or home dir */
-			char_u	*buf = alloc(1024);
-			if (!buf)
-			{
-				emsg(e_outofmem);
+			char_u	*buf = alloc(MAXPATHL);
+			if (buf == NULL)
 				return FAIL;
-			}
-			expand_env(*pat, buf, 1024);
+			expand_env(*pat, buf, MAXPATHL);
 			if (mch_has_wildcard(buf))	/* still wildcards in there? */
-			{
 				expl_files = (char_u **)_fnexplode(buf);
-			}
 			if (expl_files == NULL)
 			{
 				/*
@@ -2508,8 +2495,11 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 				 * notexist* [new file]
 				 */
 				expl_files = (char_u **)alloc(sizeof(char_u **) * 2);
-				expl_files[0] = strsave(buf);
-				expl_files[1] = NULL;
+				if (expl_files != NULL)
+				{
+					expl_files[0] = strsave(buf);
+					expl_files[1] = NULL;
+				}
 			}
 			vim_free(buf);
 		}
@@ -2520,8 +2510,11 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 			{
 				/* see above for explanation */
 				expl_files = (char_u **)alloc(sizeof(char_u **) * 2);
-				expl_files[0] = strsave(*pat);
-				expl_files[1] = NULL;
+				if (expl_files != NULL)
+				{
+					expl_files[0] = strsave(*pat);
+					expl_files[1] = NULL;
+				}
 			}
 		}
 		if (!expl_files)
@@ -2530,14 +2523,14 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 			char_u msg[128];
 			sprintf(msg, "%s (unix.c:%d)", e_internal, __LINE__);
 			emsg(msg);
-			*file = (char_u **)"";
+			*file = NULL;
 			*num_file = 0;
-			return OK;
+			return FAIL;
 		}
 		/*
 		 * Count number of names resulting from expansion,
-		 * At the same time add a backslash to the end of names that happen to be
-		 * directories, and replace slashes with backslashes.
+		 * At the same time add a backslash to the end of names that happen to
+		 * be directories, and replace slashes with backslashes.
 		 */
 		for (i = 0; (p = expl_files[i]) != NULL; i++, (*num_file)++)
 		{
@@ -2547,36 +2540,45 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 				files_alloced += EXPL_ALLOC_INC;
 				*file = (char_u **) realloc(*file,
 											sizeof(char_u **) * files_alloced);
+				if (*file == NULL)
+				{
+					emsg(e_outofmem);
+					*num_file = 0;
+					return FAIL;
+				}
 				files_free = EXPL_ALLOC_INC;
 			}
 			slash_adjust(p);
 			if (mch_isdir(p))
 			{
 				len = strlen(p);
-				p = realloc(p, len + 2);
-				if (!p)
+				if (((*file)[*num_file] = alloc(len + 2)) != NULL)
 				{
-					emsg(e_outofmem);
-					return FAIL;
+					strcpy((*file)[*num_file], p);
+					(*file)[*num_file][len] = '\\';
+					(*file)[*num_file][len+1] = 0;
 				}
-				(*file)[*num_file] = p;
-				p += len;
-				*p++ = '\\';
-				*p = 0;
 			}
 			else
 			{
 				(*file)[*num_file] = strsave(p);
 			}
+
+			/*
+			 * Error message already given by either alloc or strsave.
+			 * Should return FAIL, but returning OK works also.
+			 */
+			if ((*file)[*num_file] == NULL)
+				break;
 		}
-		_fnexplodefree(expl_files);
+		_fnexplodefree((char **)expl_files);
 	}
 	return OK;
 
 #else /* __EMX__ */
 
 	int		dir;
-	char_u	tmpname[TMPNAMELEN];
+	char_u	*tempname;
 	char_u	*command;
 	FILE	*fd;
 	char_u	*buffer;
@@ -2606,8 +2608,7 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 /*
  * get a name for the temp file
  */
-	STRCPY(tmpname, TMPNAME2);
-	if (*mktemp((char *)tmpname) == NUL)
+	if ((tempname = vim_tempname('o')) == NULL)
 	{
 		emsg(e_notmp);
 	    return FAIL;
@@ -2620,17 +2621,20 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 	if ((len = STRLEN(p_sh)) >= 3 && STRCMP(p_sh + len - 3, "csh") == 0)
 		use_glob = TRUE;
 
-	len = TMPNAMELEN + 11;
+	len = STRLEN(tempname) + 12;
 	for (i = 0; i < num_pat; ++i)		/* count the length of the patterns */
 		len += STRLEN(pat[i]) + 3;
 	command = alloc(len);
 	if (command == NULL)
+	{
+		vim_free(tempname);
 		return FAIL;
+	}
 	if (use_glob)
 		STRCPY(command, "glob >");		/* build the shell command */
 	else
 		STRCPY(command, "echo >");		/* build the shell command */
-	STRCAT(command, tmpname);
+	STRCAT(command, tempname);
 	for (i = 0; i < num_pat; ++i)
 	{
 #ifdef USE_SYSTEM
@@ -2657,7 +2661,8 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 	vim_free(command);
 	if (i == FAIL)							/* call_shell failed */
 	{
-		vim_remove(tmpname);
+		vim_remove(tempname);
+		vim_free(tempname);
 		/*
 		 * With interactive completion, the error message is not printed.
 		 * However with USE_SYSTEM, I don't know how to turn off error messages
@@ -2677,10 +2682,11 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 /*
  * read the names from the file into memory
  */
- 	fd = fopen((char *)tmpname, "r");
+ 	fd = fopen((char *)tempname, "r");
 	if (fd == NULL)
 	{
-		emsg2(e_notopen, tmpname);
+		emsg2(e_notopen, tempname);
+		vim_free(tempname);
 		return FAIL;
 	}
 	fseek(fd, 0L, SEEK_END);
@@ -2689,19 +2695,22 @@ ExpandWildCards(num_pat, pat, num_file, file, files_only, list_notfound)
 	buffer = alloc(len + 1);
 	if (buffer == NULL)
 	{
-		vim_remove(tmpname);
+		vim_remove(tempname);
+		vim_free(tempname);
 		fclose(fd);
 		return FAIL;
 	}
 	i = fread((char *)buffer, 1, len, fd);
 	fclose(fd);
-	vim_remove(tmpname);
+	vim_remove(tempname);
 	if (i != len)
 	{
-		emsg2(e_notread, tmpname);
+		emsg2(e_notread, tempname);
+		vim_free(tempname);
 		vim_free(buffer);
 		return FAIL;
 	}
+	vim_free(tempname);
 
 	if (use_glob)		/* file names are separated with NUL */
 	{
