@@ -237,7 +237,8 @@ new_win:
 				{
 					setpcmark();
 					if (win_split(0, FALSE) == OK)
-						(void)do_ecmd(0, ptr, NULL, NULL, p_hid, (linenr_t)0, FALSE);
+						(void)do_ecmd(0, ptr, NULL, NULL, (linenr_t)0,
+													   p_hid ? ECMD_HIDE : 0);
 					vim_free(ptr);
 				}
 				break;
@@ -440,6 +441,21 @@ win_split(new_height, redraw)
 	return OK;
 }
 
+/*
+ * Check if "win" is a pointer to an existing window.
+ */
+	int
+win_valid(win)
+	WIN		*win;
+{
+	WIN 	*wp;
+
+	for (wp = firstwin; wp != NULL; wp = wp->w_next)
+		if (wp == win)
+			return TRUE;
+	return FALSE;
+}
+
 /*
  * Return the number of windows.
  */
@@ -489,6 +505,14 @@ make_windows(count)
 		win_new_height(curwin, curwin->w_height - STATUS_HEIGHT);
 	}
 
+#ifdef AUTOCMD
+/*
+ * Don't execute autocommands while creating the windows.  Must do that
+ * when putting the buffers in the windows.
+ */
+	++autocmd_busy;
+#endif
+
 /*
  * set 'splitbelow' off for a moment, don't want that now
  */
@@ -501,6 +525,10 @@ make_windows(count)
 			break;
 	p_sb = p_sb_save;
 
+#ifdef AUTOCMD
+	--autocmd_busy;
+#endif
+
 		/* return actual number of windows */
 	return (count - todo);
 }
@@ -749,6 +777,9 @@ close_window(win, free_buf)
 	int		free_buf;
 {
 	WIN 	*wp;
+#ifdef AUTOCMD
+	int		other_buffer = FALSE;
+#endif
 
 	if (lastwin == firstwin)
 	{
@@ -756,6 +787,34 @@ close_window(win, free_buf)
 		return;
 	}
 
+#ifdef AUTOCMD
+	if (win == curwin)
+	{
+		/*
+		 * Guess which window is going to be the new current window.
+		 * This may change because of the autocommands (sigh).
+		 */
+		if ((!p_sb && win->w_next != NULL) || win->w_prev == NULL)
+			wp = win->w_next;
+		else
+			wp = win->w_prev;
+
+		/*
+		 * Be careful: If autocommands delete the window, return now.
+		 */
+		if (wp->w_buffer != curbuf)
+		{
+			other_buffer = TRUE;
+			apply_autocmds(EVENT_BUFLEAVE, NULL, NULL);
+			if (!win_valid(win))
+				return;
+		}
+		apply_autocmds(EVENT_WINLEAVE, NULL, NULL);
+		if (!win_valid(win))
+			return;
+	}
+#endif
+
 /*
  * Remove the window.
  * if 'splitbelow' the free space goes to the window above it.
@@ -773,15 +832,6 @@ close_window(win, free_buf)
 		wp = win->w_prev;
 	win_new_height(wp, wp->w_height + win->w_height + win->w_status_height);
 
-#ifdef AUTOCMD
-	if (win == curwin)
-	{
-		if (wp->w_buffer != curbuf)
-			apply_autocmds(EVENT_BUFLEAVE, NULL, NULL);
-		apply_autocmds(EVENT_WINLEAVE, NULL, NULL);
-	}
-#endif
-
 /*
  * Close the link to the buffer.
  */
@@ -793,7 +843,14 @@ close_window(win, free_buf)
 	if (p_ea)
 		win_equal(wp, FALSE);
 	if (curwin == NULL)
+	{
 		win_enter(wp, FALSE);
+#ifdef AUTOCMD
+		if (other_buffer)
+			apply_autocmds(EVENT_BUFENTER, NULL, NULL);
+#endif
+	}
+
 	/*
 	 * if last window has status line now and we don't want one,
 	 * remove the status line
@@ -807,7 +864,7 @@ close_window(win, free_buf)
 	}
 
 	updateScreen(NOT_VALID);
-	if (RedrawingDisabled)
+	if (RedrawingDisabled && win_valid(wp))
 		comp_Botline(wp);			/* need to do this before cursupdate() */
 }
 
@@ -888,7 +945,8 @@ win_init(wp)
 }
 
 /*
- * make window wp the current window
+ * Make window wp the current window.
+ * Can be called when curwin == NULL, if curwin already has been closed.
  */
 	void
 win_enter(wp, undo_sync)
@@ -905,12 +963,19 @@ win_enter(wp, undo_sync)
 #ifdef AUTOCMD
 	if (curwin != NULL)
 	{
+		/*
+		 * Be careful: If autocommands delete the window, return now.
+		 */
 		if (wp->w_buffer != curbuf)
 		{
 			apply_autocmds(EVENT_BUFLEAVE, NULL, NULL);
 			other_buffer = TRUE;
+			if (!win_valid(wp))
+				return;
 		}
 		apply_autocmds(EVENT_WINLEAVE, NULL, NULL);
+		if (!win_valid(wp))
+			return;
 	}
 #endif
 
@@ -919,7 +984,7 @@ win_enter(wp, undo_sync)
 		u_sync();
 		/* may have to copy the buffer options when 'cpo' contains 'S' */
 	if (wp->w_buffer != curbuf)
-		buf_copy_options(curbuf, wp->w_buffer, TRUE);
+		buf_copy_options(curbuf, wp->w_buffer, TRUE, FALSE);
 	if (curwin != NULL)
 		prevwin = curwin;		/* remember for CTRL-W p */
 	curwin = wp;
@@ -1506,7 +1571,8 @@ get_file_name_in_path(ptr, col, options)
 			if (len != 0)
 			{
 								/* Look for file relative to current file */
-				if (file_name[0] == '.' && curr_path_len > 0)
+				if (file_name[0] == '.' && curr_path_len > 0 &&
+										(len == 1 || ispathsep(file_name[1])))
 				{
 					if (len == 1)		/* just a "." */
 						len = 0;
